@@ -39,8 +39,8 @@ import tempfile
 import textwrap
 
 import charliecloud as ch
-import image as im
 import filesystem as fs
+import image as im
 import pull
 
 
@@ -263,52 +263,26 @@ class File_Metadata:
    #   st ............ Stat object for the file. Absent after un-pickling.
 
    def __init__(self, image_root, path):
+      # Note: Constructor not called during unpickle.
       self.image_root = image_root
       self.path = path
       self.path_abs = image_root // path
-      self.st = self.path_abs.stat_(False)
-      # Note: Constructor not called during unpickle.
-      for attr in ("atime_ns", "mtime_ns", "mode", "size"):
-         setattr(self, attr, getattr(self.st, "st_" + attr))
+      self.st = self.path_abs.stat(False)
+      self.stat_cache_update()
       self.children = dict()
       self.dont_restore = False
       self.hardlink_to = None
       self.large_name = None
       self.xattrs = dict()
-      if ch.save_xattrs:
-         for xattr in ch.ossafe(os.listxattr,
-                                "can’t list xattrs: %s" % self.path_abs,
-                                self.path_abs, follow_symlinks=False):
+      if ch.xattrs_save:
+         for xattr in ch.ossafe("can’t list xattrs: %s" % self.path_abs,
+                                os.listxattr, self.path_abs,
+                                follow_symlinks=False):
             self.xattrs[xattr] = \
-               ch.ossafe(os.getxattr, ("can’t get xattr: %s: %s"
+               ch.ossafe(("can’t get xattr: %s: %s"
                                        % (self.path_abs, xattr)),
-                         self.path_abs, xattr, follow_symlinks=False)
-
-   def __getstate__(self):
-      return { a:v for (a,v) in self.__dict__.items()
-                   if (a not in { "image_root", "path", "path_abs", "st" }) }
-
-   @property
-   def empty_dir_p(self):
-      """True if I represent either an empty directory, or a directory that
-         contains only children where empty_dir_p is true. E.g., the root of a
-         directory tree containing only empty directories returns true."""
-      # In principle this could do a lot of recursion, but in practice I’m
-      # guessing it’s not too much.
-      if (not stat.S_ISDIR(self.mode)):
-         return False  # not a directory
-      # True if no children (truly empty directory), or each child is unstored
-      # or empty_dir_p (recursively empty directory tree).
-      return all((c.unstored or c.empty_dir_p) for c in self.children.values())
-
-   @property
-   def unstored(self):
-      """True if I represent something not stored, either ignored by Git or
-         deleted by us before committing."""
-      return (   stat.S_ISFIFO(self.mode)
-              or stat.S_ISSOCK(self.mode)
-              or self.large_name is not None
-              or self.hardlink_to is not None)
+                         os.getxattr, self.path_abs, xattr,
+                         follow_symlinks=False)
 
    @classmethod
    def git_prepare(class_, image_root, large_file_thresh,
@@ -376,9 +350,11 @@ class File_Metadata:
          # skip Git stuff at image root
          fm.dont_restore = True
          return fm
-      # Ensure minimum permissions. Some tools like to make files with mode
-      # 000, because root ignores the permissions bits.
-      fm.path_abs.chmod_min(fm.st)
+      # Ensure minimum permissions. Some tools like to make files without
+      # necessary owner permissions, because root ignores the permissions bits
+      # (CAP_DAC_OVERRIDE). See e.g. #1765.
+      fm.st = fm.path_abs.chmod_min(fm.st)
+      fm.stat_cache_update()
       # Validate file type and recurse if needed. (Don’t use os.walk() because
       # it’s iterative, and our algorithm is better expressed recursively.)
       if   (   stat.S_ISREG(fm.mode)
@@ -413,7 +389,7 @@ class File_Metadata:
             ch.DEBUG("hard link: deleting subsequent: %d %d %s"
                      % (fm.st.st_dev, fm.st.st_ino, path))
             fm.hardlink_to = hardlinks[(fm.st.st_dev, fm.st.st_ino)]
-            fm.path_abs.unlink_()
+            fm.path_abs.unlink()
             return fm
          else:
             ch.DEBUG("hard link: recording first: %d %d %s"
@@ -432,16 +408,16 @@ class File_Metadata:
       # Remove empty directories. Git will ignore them, including leaving them
       # there when switching the worktree to a different branch, which is bad.
       if (fm.empty_dir_p):
-         fm.path_abs.rmdir_()
+         fm.path_abs.rmdir()
          return fm
       # Remove FIFOs for the same reason.
       if (stat.S_ISFIFO(fm.mode)):
          fm.path_abs.unlink()
          return fm
       # Rename if necessary.
-      if (path.git_incompatible_p):
+      if (not path.git_compatible_p):
          ch.DEBUG("renaming: %s -> %s" % (path, path.git_escaped))
-         fm.path_abs.rename_(fm.path_abs.git_escaped)
+         fm.path_abs.rename(fm.path_abs.git_escaped)
       # Done.
       return fm
 
@@ -453,6 +429,32 @@ class File_Metadata:
       fm_tree.unpickle_fix(image_root, path=fs.Path("."))
       return fm_tree
 
+   def __getstate__(self):
+      return { a:v for (a,v) in self.__dict__.items()
+                   if (a not in { "image_root", "path", "path_abs", "st" }) }
+
+   @property
+   def empty_dir_p(self):
+      """True if I represent either an empty directory, or a directory that
+         contains only children where empty_dir_p is true. E.g., the root of a
+         directory tree containing only empty directories returns true."""
+      # In principle this could do a lot of recursion, but in practice I’m
+      # guessing it’s not too much.
+      if (not stat.S_ISDIR(self.mode)):
+         return False  # not a directory
+      # True if no children (truly empty directory), or each child is unstored
+      # or empty_dir_p (recursively empty directory tree).
+      return all((c.unstored or c.empty_dir_p) for c in self.children.values())
+
+   @property
+   def unstored(self):
+      """True if I represent something not stored, either ignored by Git or
+         deleted by us before committing."""
+      return (   stat.S_ISFIFO(self.mode)
+              or stat.S_ISSOCK(self.mode)
+              or self.large_name is not None
+              or self.hardlink_to is not None)
+
    def get(self, path):
       "Return the File_Metadata object at path."
       fm = self
@@ -475,19 +477,19 @@ class File_Metadata:
          # to subsequent (unstored) links.
          target = self.image_root // self.hardlink_to
          ch.DEBUG("hard link: restoring: %s -> %s" % (self.path_abs, target))
-         ch.ossafe(os.link, "can’t hardlink: %s -> %s" % (self.path_abs,
-                                                          target),
-                   target, self.path_abs, follow_symlinks=False)
+         ch.ossafe("can’t hardlink: %s -> %s" % (self.path_abs,
+                                                       target),
+                   os.link, target, self.path_abs, follow_symlinks=False)
       elif (self.large_name is not None):
          self.large_restore()
       elif (self.empty_dir_p):
-         ch.ossafe(os.mkdir, "can’t mkdir: %s" % self.path, self.path_abs)
+         ch.ossafe("can’t mkdir: %s" % self.path, os.mkdir, self.path_abs)
       elif (stat.S_ISFIFO(self.mode)):
-         ch.ossafe(os.mkfifo, "can’t make FIFO: %s" % self.path, self.path_abs)
-      elif (self.path.git_incompatible_p):
-         self.path_abs.git_escaped.rename_(self.path_abs)
+         ch.ossafe("can’t make FIFO: %s" % self.path, os.mkfifo, self.path_abs)
+      elif (not self.path.git_compatible_p):
+         self.path_abs.git_escaped.rename(self.path_abs)
       for (xattr, val) in self.xattrs.items():
-         self.path_abs.setxattr(xattr, val, follow_symlinks=False)
+         self.path_abs.setxattr(xattr, val)
       # Recurse children.
       if (len(self.children) > 0):
          for child in self.children.values():
@@ -498,9 +500,9 @@ class File_Metadata:
            or stat.S_ISDIR(self.mode)        # maybe just created or modified
            or stat.S_ISFIFO(self.mode))      # we just made the FIFO
           and not stat.S_ISLNK(self.mode)):  # can’t not follow symlinks
-         ch.ossafe(os.utime, "can’t restore times: %s" % self.path_abs,
+         ch.ossafe("can’t restore times: %s" % self.path_abs, os.utime,
                    self.path_abs, ns=(self.atime_ns, self.mtime_ns))
-         ch.ossafe(os.chmod, "can’t restore mode: %s" % self.path_abs,
+         ch.ossafe("can’t restore mode: %s" % self.path_abs, os.chmod,
                    self.path_abs, stat.S_IMODE(self.mode))
 
    def large_name_get(self):
@@ -528,12 +530,12 @@ class File_Metadata:
          exists, then return the appropriate large name."""
       large_name = self.large_name_get()
       target = ch.storage.build_large_path(large_name)
-      if (target.exists_()):
+      if (target.exists()):
          op = "found"
-         self.path_abs.unlink_()
+         self.path_abs.unlink()
       else:
          op = "moving to"
-         self.path_abs.rename_(target)
+         self.path_abs.rename(target)
       ch.DEBUG("large file: %s: %s: %s" % (self.path, op, large_name))
       return large_name
 
@@ -547,6 +549,10 @@ class File_Metadata:
       (self.image_root // PICKLE_PATH) \
          .file_write(pickle.dumps(self, protocol=4))
 
+   def stat_cache_update(self):
+      for attr in ("atime_ns", "mtime_ns", "mode", "size"):
+         setattr(self, attr, getattr(self.st, "st_" + attr))
+
    def str_for_log(self):
       # Truncate reported time to seconds.
       fmt = "%Y-%m-%d.%H:%M:%S"
@@ -634,10 +640,6 @@ class State_ID:
          ch.FATAL("state ID: malformed hex: %s" % text);
       return class_(b)
 
-   @property
-   def short(self):
-      return str(self)[:4]
-
    def __eq__(self, other):
       return self.id_ == other.id_
 
@@ -648,12 +650,16 @@ class State_ID:
       s = self.id_.hex().upper()
       return ":".join((s[0:4], s[4:8], s[8:16], s[16:24], s[24:32]))
 
+   @property
+   def short(self):
+      return str(self)[:4]
 
-## Main classes ##
+
+## Core classes ##
 
 class Enabled_Cache:
 
-   root_id =   State_ID.from_text("4A6F:73C3:A9204361:7061626C:616E6361")
+   root_id = State_ID.from_text("4A6F:73C3:A9204361:7061626C:616E6361")
 
    __slots__ = ("bootstrap_ct",
                 "file_metadata",
@@ -675,13 +681,6 @@ class Enabled_Cache:
          self.configure()         # updates config if needed
       self.worktrees_fix()
 
-   def __str__(self):
-      return ("enabled (large=%g)" % self.large_threshold)
-
-   @property
-   def root(self):
-      return ch.storage.build_cache
-
    @staticmethod
    def branch_name_ready(ref):
       return ref.for_path
@@ -698,6 +697,13 @@ class Enabled_Cache:
          branch/tag names that look like hashes."""
       return (re.search(r"^[0-9a-f]{7,}$", commit_ish) is not None)
 
+   def __str__(self):
+      return ("enabled (large=%g)" % self.large_threshold)
+
+   @property
+   def root(self):
+      return ch.storage.build_cache
+
    def adopt(self, img):
       self.worktree_adopt(img, "root")
       img.metadata_load()
@@ -833,7 +839,7 @@ class Enabled_Cache:
    def configure(self):
       # Configuration.
       path = self.root // "config"
-      fp = path.open_("r+")
+      fp = path.open("r+")
       config = configparser.ConfigParser()
       config.read_file(fp, source=path)
       changed = False
@@ -850,7 +856,7 @@ class Enabled_Cache:
          ch.VERBOSE("writing updated Git config")
          fp.seek(0)
          fp.truncate()
-         ch.ossafe(config.write, "can’t write Git config: %s" % path, fp)
+         ch.ossafe("can’t write Git config: %s" % path, config.write, fp)
       ch.close_(fp)
       # Ignore list entries:
       #
@@ -875,7 +881,7 @@ class Enabled_Cache:
       # to prevent deleted .gitignore files from creeping back in. (I couldn’t
       # figure out how to fix this for “filter-branch” either, which I didn’t
       # want to use anyway for the reasons above.)
-      if (ch.storage.bucache_needs_ignore_upgrade.exists_()):
+      if (ch.storage.bucache_needs_ignore_upgrade.exists()):
          ch.INFO("upgrading build cache to v6+, some cached data may be lost",
                  "see release notes for v0.32")
          text = self.git(["fast-export", "--no-data", "--", "--all"],
@@ -901,7 +907,7 @@ class Enabled_Cache:
             i += 1
          text = "\n".join(lines)
          text = re.sub(r"^(D|M [0-7]+ [0-9a-f]+) \.(git|weirdal_)ignore$",
-                       "#\g<0>", text, flags=re.MULTILINE)
+                       r"#\g<0>", text, flags=re.MULTILINE)
          #fs.Path("/tmp/new").file_write(text)
          self.git(["fast-import", "--force"], input=text)
          self.git(["reflog", "expire", "--all", "--expire=now"])
@@ -986,7 +992,7 @@ class Enabled_Cache:
       ch.INFO("found %d large files used; deleting others" % len(larges_used))
       for l in ch.storage.build_large.listdir():
          if (l not in larges_used):
-            (ch.storage.build_large // l).unlink_()
+            (ch.storage.build_large // l).unlink()
       t.log("deleted unused large files")
 
    def git(self, argv, cwd=None, quiet=True, *args, **kwargs):
@@ -1100,7 +1106,7 @@ class Enabled_Cache:
          pid_path = ch.storage.build_cache // "gc.pid"
          try:
             fp = open(pid_path, "rt", encoding="UTF-8")
-            text = ch.ossafe(fp.read, "can’t read: %s" % pid_path)
+            text = ch.ossafe("can’t read: %s" % pid_path, fp.read)
             pid = int(text.split()[0])
             ch.INFO("stopping build cache garbage collection, PID %d" % pid)
             ch.kill_blocking(pid)
@@ -1264,8 +1270,8 @@ class Enabled_Cache:
          # Move GIT_DIR from default location to where we want it.
          git_dir_default = image.unpack_path // ".git"
          git_dir_new = image.unpack_path // im.GIT_DIR
-         git_dir_new.parent.mkdir_()
-         git_dir_default.rename_(git_dir_new)
+         git_dir_new.parent.mkdir()
+         git_dir_default.rename(git_dir_new)
          t.log("created worktree")
 
    def worktree_adopt(self, image, base):
@@ -1278,12 +1284,12 @@ class Enabled_Cache:
          ch.WARNING("temporary image still exists, deleting",
                     "maybe a previous command crashed?")
          ch.storage.image_tmp.rmtree()
-      image.unpack_path.rename_(ch.storage.image_tmp)
+      image.unpack_path.rename(ch.storage.image_tmp)
       self.worktree_add(image, base)
       (image.unpack_path // im.GIT_DIR).rename(   ch.storage.image_tmp
                                                // im.GIT_DIR)
       image.unpack_path.rmtree()
-      ch.storage.image_tmp.rename_(image.unpack_path)
+      ch.storage.image_tmp.rename(image.unpack_path)
 
    def worktree_head(self, image):
       cp = self.git(["rev-parse", "--short", "HEAD"],
@@ -1318,6 +1324,15 @@ class Enabled_Cache:
                                              // "*" // im.GIT_DIR)) }
       wt_gits =    { fs.Path(i).name
                      for i in glob.iglob("%s/worktrees/*" % self.root) }
+      # Unlink images that think they are in Git but are not. This should not
+      # happen, but it does, and I wasn’t able to figure out how it happened.
+      wt_gits_orphaned = wt_actuals - wt_gits
+      for img_dir in wt_gits_orphaned:
+         link = ch.storage.unpack_base // img_dir // im.GIT_DIR
+         ch.WARNING("image erroneously marked cached, fixing: %s" % link,
+                    ch.BUG_REPORT_PLZ)
+         link.unlink()
+      wt_actuals -= wt_gits_orphaned
       # Delete worktree data for images that no longer exist or aren’t
       # Git-enabled any more.
       wt_gits_deleted = wt_gits - wt_actuals
@@ -1325,7 +1340,12 @@ class Enabled_Cache:
          (ch.storage.build_cache // "worktrees" // wt).rmtree()
       ch.VERBOSE("deleted %d stale worktree metadatas" % len(wt_gits_deleted))
       wt_gits -= wt_gits_deleted
-      assert (wt_gits == wt_actuals)
+      # Validate that the pointers are in sync now.
+      if (wt_gits != wt_actuals):
+         ch.ERROR("found images -> cache links: %s" % " ".join(wt_actuals))
+         ch.ERROR("found cache -> images links: %s" % " ".join(wt_gits))
+         ch.FATAL("build cache is desynchronized, cannot proceed",
+                  ch.BUG_REPORT_PLZ)
       # If storage directory moved, repair all the paths.
       if (len(wt_gits) > 0):
          wt_dir_stored = fs.Path((   ch.storage.build_cache
