@@ -1,8 +1,8 @@
 import argparse
 import atexit
+import cProfile
 import collections
 import collections.abc
-import cProfile
 import datetime
 import enum
 import functools
@@ -20,11 +20,12 @@ import subprocess
 import sys
 import time
 import traceback
+import warnings
 
 
 # List of dependency problems. This variable needs to be created before we
 # import any other Charliecloud stuff to avoid #806.
-depfails = []
+depfails = []  # 👻
 
 
 import filesystem as fs
@@ -103,6 +104,10 @@ ARCH_MAP = { "armv5l":    "arm/v5",
 ARCH_MAP_FALLBACK = { "arm/v7": ("arm",),
                       "arm64/v8": ("arm64",) }
 
+# Incompatible option pairs for the ch-image command line
+CLI_INCOMPATIBLE_OPTS = [("quiet", "verbose"),
+                         ("xattrs", "no_xattrs")]
+
 # String to use as hint when we throw an error that suggests a bug.
 BUG_REPORT_PLZ = "please report this bug: https://github.com/hpc/charliecloud/issues"
 
@@ -145,7 +150,7 @@ log_fp = sys.stderr       # File object to print logs to.
 trace_fatal = False       # Add abbreviated traceback to fatal error hint.
 
 # Warnings to be re-printed when program exits
-warnings = list()
+warns = list()
 
 # True if the download cache is enabled.
 dlcache_p = None
@@ -154,6 +159,9 @@ dlcache_p = None
 profiling = False
 profile = None
 
+# Width of terminal.
+term_width = shutil.get_terminal_size(fallback=(sys.maxsize, -1))[0]
+
 
 ## Exceptions ##
 
@@ -379,7 +387,7 @@ class Progress_Reader:
          close_(self.fp)
 
    def read(self, size=-1):
-     data = ossafe(self.fp.read, "can’t read: %s" % self.fp.name, size)
+     data = ossafe("can’t read: %s" % self.fp.name, self.fp.read, size)
      self.progress.update(len(data))
      return data
 
@@ -431,16 +439,16 @@ class Progress_Writer:
       if (self.progress is not None):
          self.progress.done()
          close_(self.fp)
-         self.path.unlink_(missing_ok=True)
-         self.path_tmp.rename_(self.path)
+         self.path.unlink(missing_ok=True)
+         self.path_tmp.rename(self.path)
 
    def start(self, length):
       self.progress = Progress(self.msg, "MiB", 2**20, length)
-      self.fp = self.path_tmp.open_("wb")
+      self.fp = self.path_tmp.open("wb")
 
    def write(self, data):
       self.progress.update(len(data))
-      ossafe(self.fp.write, "can’t write: %s" % self.path, data)
+      ossafe("can’t write: %s" % self.path, self.fp.write, data)
 
 
 class Timer:
@@ -474,6 +482,10 @@ def FATAL(msg, hint=None, **kwargs):
       tr = None
    raise Fatal_Error(msg, hint, tr, **kwargs)
 
+def ILLERI(msg, hint=None, **kwargs):
+   # For temporary debugging only. See contributors’ guide.
+   log(msg, hint, None, "38;5;207m", "", **kwargs)  # hot pink
+
 def INFO(msg, hint=None, **kwargs):
    "Note: Use print() for output; this function is for logging."
    if (log_level >= Log_Level.INFO):
@@ -490,7 +502,7 @@ def VERBOSE(msg, hint=None, **kwargs):
 def WARNING(msg, hint=None, msg_save=True, **kwargs):
    if (log_level > Log_Level.STDERR):
       if (msg_save):
-         warnings.append(msg)
+         warns.append(msg)
       log(msg, hint, None, "31m", "warning: ", **kwargs)  # red
 
 def arch_host_get():
@@ -528,7 +540,7 @@ def close_(fp):
       path = fp.name
    except AttributeError:
       path = "(no path)"
-   ossafe(fp.close, "can’t close: %s" % path)
+   ossafe("can’t close: %s" % path, fp.close)
 
 def cmd(argv, fail_ok=False, **kwargs):
    """Run command using cmd_base(). If fail_ok, return the exit code whether
@@ -653,12 +665,17 @@ def exit(code):
 
 def init(cli):
    # logging
-   global log_festoon, log_fp, log_level, trace_fatal, save_xattrs
-   save_xattrs = (not cli.no_xattrs)
+   global log_festoon, log_fp, log_level, trace_fatal, xattrs_save
+   incomp_opts = 0
+   for (x,y) in CLI_INCOMPATIBLE_OPTS:
+      if (getattr(cli, x) and getattr(cli, y)):
+         ERROR("“--%s” incompatible with “--%s”" % ((x.replace("_","-"),
+                                                     y.replace("_","-"))))
+         incomp_opts += 1
+   if (incomp_opts > 0):
+      FATAL("%d incompatible option pair(s)" % incomp_opts)
+   xattrs_save = ((cli.xattrs) or (("CH_XATTRS" in os.environ) and (not cli.no_xattrs)))
    trace_fatal = (cli.debug or bool(os.environ.get("CH_IMAGE_DEBUG", False)))
-   if (cli.quiet and cli.verbose):
-      ERROR("“--quiet” incompatible with “--verbose”")
-      FATAL("incompatible option")
    log_level = Log_Level(cli.verbose - cli.quiet)
    assert (-3 <= log_level.value <= 3)
    if (log_level <= Log_Level.STDERR):
@@ -668,10 +685,11 @@ def init(cli):
       log_festoon = True
    file_ = os.getenv("CH_LOG_FILE")
    if (file_ is not None):
-      log_fp = file_.open_("at")
+      log_fp = file_.open("at")
    atexit.register(color_reset, log_fp)
    VERBOSE("version: %s" % version.VERSION)
    VERBOSE("verbose level: %d (%s))" % (log_level.value, log_level.name))
+   VERBOSE("save xattrs: %s" % str(xattrs_save))
    # signal handling
    signal.signal(signal.SIGINT, sigterm)
    signal.signal(signal.SIGTERM, sigterm)
@@ -704,6 +722,12 @@ def init(cli):
    else:
       rg.auth_p = False
    VERBOSE("registry authentication: %s" % rg.auth_p)
+   # Red Hat Python warns about tar bugs, citing CVE-2007-4559.
+   # We mitigate this already, so suppress the noise. (#1818)
+   warnings.filterwarnings("ignore", module=r"^tarfile$",
+                           message=(  "^The default behavior of tarfile"
+                                    + " extraction has been changed to disallow"
+                                    + " common exploits"))
    # misc
    global password_many, profiling
    password_many = cli.password_many
@@ -776,7 +800,7 @@ def monkey_write_streams():
 def now_utc_iso8601():
    return datetime.datetime.utcnow().isoformat(timespec="seconds") + "Z"
 
-def ossafe(f, msg, *args, **kwargs):
+def ossafe(msg, f, *args, **kwargs):
    """Call f with args and kwargs. Catch OSError and other problems and fail
       with a nice error message."""
    try:
@@ -870,7 +894,7 @@ def variables_sub(s, variables):
       m = re.search(r"(?<!\\)\${.+?:[+-].+?}", s)
       if (m is not None):
          FATAL("modifiers ${foo:+bar} and ${foo:-bar} not yet supported (issue #774)")
-      s = re.sub(r"(?<!\\)\${?%s}?" % k, v, s)
+      s = re.sub(r"(?<!\\)\$({%s}|%s(?=\W|$))" % (k, k), v, s)
    return s
 
 def version_check(argv, min_, required=True, regex=r"(\d+)\.(\d+)\.(\d+)"):
@@ -913,14 +937,14 @@ def walk(*args, **kwargs):
       3-tuple (dirpath, dirnames, filenames), where dirpath is a Path object,
       and dirnames and filenames are lists of Path objects. For insight into
       these being lists rather than generators, see use of ch.walk() in
-      I_copy.copy_src_dir()."""
+      Copy_G.copy_src_dir()."""
    for (dirpath, dirnames, filenames) in os.walk(*args, **kwargs):
       yield (fs.Path(dirpath),
              [fs.Path(dirname) for dirname in dirnames],
              [fs.Path(filename) for filename in filenames])
 
 def warnings_dump():
-   if (len(warnings) > 0):
-      WARNING("reprinting %d warning(s)" % len(warnings), msg_save=False)
-   for msg in warnings:
+   if (len(warns) > 0):
+      WARNING("reprinting %d warning(s)" % len(warns), msg_save=False)
+   for msg in warns:
       WARNING(msg, msg_save=False)
