@@ -183,8 +183,8 @@ FAKEROOT_DEFAULT_CONFIGS = {
    { "name": "RHEL 7 and derivatives",
      "match": ("/etc/redhat-release", r"release 7\."),
      "init": [ ("command -v fakeroot > /dev/null",
-                "set -ex; "
-                "if ! grep -Eq '\[epel\]' /etc/yum.conf /etc/yum.repos.d/*; then "
+                "set -e; "
+                r"if ! grep -Eq '\[epel\]' /etc/yum.conf /etc/yum.repos.d/*; then "
                 "yum install -y https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm; "
                 "yum install -y fakeroot; "
                 "yum remove -y epel-release; "
@@ -198,8 +198,8 @@ FAKEROOT_DEFAULT_CONFIGS = {
    { "name": "RHEL 8+ and derivatives",
      "match":  ("/etc/redhat-release", r"release (?![0-7]\.)"),
      "init": [ ("command -v fakeroot > /dev/null",
-                "set -ex; "
-                "if ! grep -Eq '\[epel\]' /etc/yum.conf /etc/yum.repos.d/*; then "
+                "set -e; "
+                r"if ! grep -Eq '\[epel\]' /etc/yum.conf /etc/yum.repos.d/*; then "
                 # Macro %rhel from *-release* RPM, e.g. redhat-release-server
                 # or centos-linux-release; thus reliable.
                 "dnf install -y https://dl.fedoraproject.org/pub/epel/epel-release-latest-$(rpm -E %rhel).noarch.rpm; "
