@@ -75,7 +75,7 @@ _string_list: "[" _WS? STRING_QUOTED ( "," _WS? STRING_QUOTED )* _WS? "]"
 _WSH: /[ \t]/+                   // sequence of horizontal whitespace
 _LINE_CONTINUE: "\\" _WSH? "\n"  // line continuation
 _WS: ( _WSH | _LINE_CONTINUE )+  // horizontal whitespace w/ line continuations
-_NEWLINES: ( _WSH? "\n" )+       // sequence of newlines
+_NEWLINES: ( _WS? "\n" )+        // sequence of newlines
 
 %import common.ESCAPED_STRING -> STRING_QUOTED
 """
@@ -120,7 +120,7 @@ env_space: WORD _WS _line
 env_equalses: env_equals ( _WS env_equals )*
 env_equals: WORD "=" ( WORD | STRING_QUOTED )
 
-from_: "FROM"i ( _WS ( option | option_keypair ) )* _WS image_ref [ _WS from_alias ] _NEWLINES
+from_: "FROM"i ( _WS ( option | option_keypair ) )* _WS image_ref ( _WS from_alias )? _NEWLINES
 from_alias: "AS"i _WS IR_PATH_COMPONENT  // FIXME: undocumented; this is guess
 
 label: "LABEL"i _WS ( label_space | label_equalses ) _NEWLINES
@@ -155,7 +155,7 @@ OPTION_VALUE: /[^= \t\n]+/
 OPTION_VAR: /[a-z]+/
 
 image_ref: IMAGE_REF
-IMAGE_REF: /[A-Za-z0-9$:._\/-]+/
+IMAGE_REF: /[${}A-Za-z0-9:._\/-]+/  // variable substitution chars ${} added
 """ + GRAMMAR_COMMON
 
 # Grammar for image references.
@@ -181,6 +181,9 @@ IR_TAG: /[A-Za-z0-9_.-]+/
 # Top-level directories we create if not present.
 STANDARD_DIRS = { "bin", "dev", "etc", "mnt", "proc", "sys", "tmp", "usr" }
 
+# Width of token name when truncating text to fit on screen.
+WIDTH_TOKEN_MAX = 10
+
 
 ## Classes ##
 
@@ -210,6 +213,16 @@ class Image:
          self.unpack_path = ch.storage.unpack(self.ref)
       self.metadata_init()
 
+   @classmethod
+   def glob(class_, image_glob):
+      """Return a possibly-empty iterator of images in the storage directory
+         matching the given glob."""
+      for ref in Reference.glob(image_glob):
+         yield class_(ref)
+
+   def __str__(self):
+      return str(self.ref)
+
    @property
    def deleteable(self):
       """True if it’s OK to delete me, either my unpack directory (a) is at
@@ -228,7 +241,7 @@ class Image:
       # Return the last modified time of self as a datetime.datetime object in
       # the local time zone.
       return datetime.datetime.fromtimestamp(
-                 (self.metadata_path // "metadata.json").stat_(False).st_mtime,
+                 (self.metadata_path // "metadata.json").stat(False).st_mtime,
                  datetime.timezone.utc).astimezone()
 
    @property
@@ -237,22 +250,12 @@ class Image:
 
    @property
    def unpack_cache_linked(self):
-      return (self.unpack_path // GIT_DIR).exists_()
+      return (self.unpack_path // GIT_DIR).exists()
 
    @property
    def unpack_exist_p(self):
       return os.path.exists(self.unpack_path)
 
-   def __str__(self):
-      return str(self.ref)
-
-   @classmethod
-   def glob(class_, image_glob):
-      """Return a possibly-empty iterator of images in the storage directory
-         matching the given glob."""
-      for ref in Reference.glob(image_glob):
-         yield class_(ref)
-
    def commit(self):
       "Commit the current unpack directory into the layer cache."
       assert False, "unimplemented"
@@ -269,7 +272,7 @@ class Image:
       ch.VERBOSE("copying image: %s -> %s" % (src_path, self.unpack_path))
       fs.Path(src_path).copytree(self.unpack_path, symlinks=True)
       # Simpler to copy this file then delete it, rather than filter it out.
-      (self.unpack_path // GIT_DIR).unlink_(missing_ok=True)
+      (self.unpack_path // GIT_DIR).unlink(missing_ok=True)
       self.unpack_init()
 
    def layers_open(self, layer_tars):
@@ -504,7 +507,7 @@ class Image:
          present will be left unchanged. After this, self.unpack_path is a
          valid Charliecloud image directory."""
       # Metadata directory.
-      (self.unpack_path // "ch").mkdir_()
+      (self.unpack_path // "ch").mkdir()
       (self.unpack_path // "ch/environment").file_ensure_exists()
       # Essential directories & mount points. Do nothing if something already
       # exists, without dereferencing, in case it’s a symlink, which will work
@@ -523,7 +526,7 @@ class Image:
       layers = self.layers_open(layer_tars)
       self.validate_members(layers)
       self.whiteouts_resolve(layers)
-      self.unpack_path.mkdir_()  # create directory in case no layers
+      self.unpack_path.mkdir()  # create directory in case no layers
       for (i, (lh, (fp, members))) in enumerate(layers.items(), start=1):
          lh_short = lh[:7]
          if (i > last_layer):
@@ -558,6 +561,7 @@ class Image:
             # Correct absolute paths.
             if (m.name.is_absolute()):
                m.name = m.name.relative_to("/")
+               abs_ct += 1
             # Record top-level directory.
             if (len(m.name.parts) > 1 or m.isdir()):
                top_dirs.add(m.name.first)
@@ -715,21 +719,6 @@ class Reference:
       elif (src is not None):
          assert False, "unsupported initialization type"
 
-   def __str__(self):
-      out = ""
-      if (self.host is not None):
-         out += self.host
-      if (self.port is not None):
-         out += ":" + str(self.port)
-      if (self.host is not None):
-         out += "/"
-      out += self.path_full
-      if (self.tag is not None):
-         out += ":" + self.tag
-      if (self.digest is not None):
-         out += "@sha256:" + self.digest
-      return out
-
    @staticmethod
    def path_to_ref(path):
       if (isinstance(path, fs.Path)):
@@ -771,6 +760,21 @@ class Reference:
       ch.DEBUG(tree.pretty())
       return tree
 
+   def __str__(self):
+      out = ""
+      if (self.host is not None):
+         out += self.host
+      if (self.port is not None):
+         out += ":" + str(self.port)
+      if (self.host is not None):
+         out += "/"
+      out += self.path_full
+      if (self.tag is not None):
+         out += ":" + self.tag
+      if (self.digest is not None):
+         out += "@sha256:" + self.digest
+      return out
+
    @property
    def as_verbose_str(self):
       def fmt(x):
@@ -863,6 +867,29 @@ fields:
 
 class Tree(lark.tree.Tree):
 
+   def _pretty(self, level, istr):
+      # Re-implement with less space optimization and more debugging info.
+      # See: https://github.com/lark-parser/lark/blob/262ab71/lark/tree.py#L78
+      pfx = "%4d %3d%s" % (self.meta.line, self.meta.column, istr*(level+1))
+      yield (pfx + self._pretty_label() + "\n")
+      for c in self.children:
+         if (isinstance(c, Tree)):
+            yield from c._pretty(level + 1, istr)
+         else:
+            text = c
+            type_ = c.type
+            width = len(pfx) + len(istr) + len(text) + len(type_) + 2
+            over = width - ch.term_width
+            if (len(type_) > WIDTH_TOKEN_MAX):
+               # trim token (unconditionally for consistent alignment)
+               token_rm = len(type_) - WIDTH_TOKEN_MAX
+               type_ = type_[:-token_rm]
+               over -= token_rm
+            if (over > 0):
+               # trim text (if needed)
+               text = text[:-(over + 3)] + "..."
+            yield "%s%s %s %s\n" % (pfx, istr, type_, text)
+
    def child(self, cname):
       """Locate a descendant subtree named cname using breadth-first search
          and return it. If no such subtree exists, return None."""
